package Xconfig::proprietary; # $Id: proprietary.pm 61415 2006-09-14 10:50:00Z pixel $




use common;

my $lib = arch() =~ /x86_64/ ? "lib64" : "lib";

sub install_matrox_hal {
    my ($prefix) = @_;
    my $tmpdir = "$prefix/root/tmp";

    my $tar = "mgadrivers-2.0.tgz";
    my $dir_in_tar = "mgadrivers";
    my $dest_dir = "$prefix/usr/lib/xorg/modules/drivers";

    #- already installed
    return if -e "$dest_dir/mga_hal_drv.o" || $::testing;

    system("wget -O $tmpdir/$tar ftp://ftp.matrox.com/pub/mga/archive/linux/2002/$tar") if !-e "$tmpdir/$tar";
    system("tar xzC $tmpdir -f $tmpdir/$tar");

    my $src_dir = "$tmpdir/$dir_in_tar/xfree86/4.2.1/drivers";
    foreach (all($src_dir)) {
	my $src = "$src_dir/$_";
	my $dest = "$dest_dir/$_";
	rename $dest, "$dest.non_hal";
	cp_af($src, $dest_dir);
    }
    rm_rf("$tmpdir/$tar");
    rm_rf("$tmpdir/$dir_in_tar");
}

sub pkgs_for_Driver2 {
    my ($card, $do_pkgs) = @_;

    my %pkg = (
	#- using current NVIDIA driver for recent NVIDIA cards (Geforce/Quadro 3/4/FX/6x00/NVS):
	nvidia => 'nvidia',
	#- using NVIDIA Legacy driver for old NVIDIA cards (TNT, TNT2, Vanta, Quadro, Quadro2, GeForce and GeForce2):
	nvidia_legacy => 'nvidia_legacy',
	#- using ATI fglrx driver (Radeon, Fire GL cards only).
	fglrx => 'ati',
    );
    my $wanted =     $card->{Driver2} . ($card->{NVIDIA_LEGACY} ? '_legacy' : '');
    my $not_wanted = $card->{Driver2} . ($card->{NVIDIA_LEGACY} ? '' : '_legacy');

    $pkg{$wanted} && $do_pkgs->is_available($pkg{$wanted}) or return;

    my $module_pkgs = $do_pkgs->check_kernel_module_packages($pkg{$wanted} . '-kernel') or
      log::l("$pkg{$wanted} available, but no kernel module package (for installed kernels, and no dkms)"), return;

    if ($card->{Driver2} eq 'nvidia') {
	#- Prevent having both nvidia & nvidia_legacy driver installed due to incompatibility.
	my $p = $do_pkgs->check_kernel_module_packages($pkg{$not_wanted} . '-kernel');
	if (my @to_remove = $do_pkgs->are_installed($pkg{$not_wanted}, if_($p, @$p))) {
	    log::l('Prevent having both nvidia & nvidia_legacy driver installed due to incompatibility.');
	    $do_pkgs->remove(@to_remove);
	}
    }

    ($pkg{$wanted}, @$module_pkgs);
}

sub may_use_Driver2 {
    my ($card) = @_;

    my $modules_dir = "/usr/$lib/xorg/modules";
    my $new_nvidia_libglx        = "$modules_dir/extensions/nvidia/libglx.so";
    my $new_nvidia_legacy_libglx = "$modules_dir/extensions/nvidia_legacy/libglx.so";
    my $old_nvidia_libglx = "$modules_dir/extensions/libglx.so";
    #- make sure everything is correct at this point, packages have really been installed
    #- and driver and GLX extension is present.

    my $check_drv = sub {
	my ($drv, $o_subdir) = @_;
	my @l = (if_($o_subdir, "$modules_dir/drivers/$o_subdir/$drv.so"),
		 "$modules_dir/drivers/$drv.so",
		 "$modules_dir/drivers/$drv.o");
	my $has = find { -e "$::prefix$_" } @l;
	$has or log::l("proprietary $drv driver missing, defaulting to free software driver (we searched for: @l)");
	$has;
    };

    my $card2 = { 
	%$card,
	Driver => $card->{Driver2}, 
	REMOVE_GLX => [ $old_nvidia_libglx, $new_nvidia_libglx, $new_nvidia_legacy_libglx ],
    };

    if ($card->{Driver2} eq 'nvidia') {
	$check_drv->('nvidia_drv', $card->{NVIDIA_LEGACY} ? 'nvidia_legacy' : 'nvidia') or return;

	#- when there is extensions/libglx.a, it means extensions/libglx.so is not xorg's libglx, so it may be nvidia's
	#- new nvidia packages have libglx.so in extensions/nvidia instead of simply extensions/
	my $libglx_a = -e "$::prefix$modules_dir/extensions/libglx.a";
	my $new_libglx = $card->{NVIDIA_LEGACY} ? $new_nvidia_legacy_libglx : $new_nvidia_libglx;
	my $libglx = find { -l "$::prefix$_" } $new_libglx, if_($libglx_a, $old_nvidia_libglx)
	  or log::l("special NVIDIA libglx missing, defaulting to free software driver"), return;

	log::explanations("Using specific NVIDIA driver and GLX extensions");
	$card2->{DRI_GLX_SPECIAL} = $libglx;
	$card2->{Options}{IgnoreEDID} = 1 if $card->{NVIDIA_LEGACY};
	$card2;
    } elsif ($card->{Driver2} eq 'fglrx') {
	$check_drv->('fglrx_drv') or return;
	-e "$::prefix$modules_dir/dri/fglrx_dri.so" || -e "$::prefix/usr/$lib/dri/fglrx_dri.so" or
	  log::l("proprietary fglrx_dri.so missing, defaulting to free software driver"), return;

	log::explanations("Using specific ATI fglrx and DRI drivers");
	$card2->{DRI_GLX} = 1;
	$card2;
    } else {
	undef;
    }
}

1;

